@@ -13,8 +13,9 @@ from . import (
     create_client,
     handle_request,
     options,
-    log,
+    debug,
     confirm_action,
+    set_value_from_config,
     MutuallyExclusiveOption,
     GLOBAL_OPTIONS,
     OUTPUT_FORMAT_OPTION,
@@ -25,6 +26,13 @@ from . import (
 from .ssh_key import (
     _ssh_key_list,
     _ssh_key_new,
+    _ssh_key_add,
+)
+from .image import _image_list
+from .project import (
+    _project_list,
+    _project_resource_list_servers,
+    _project_resource_move,
 )
 
 
@@ -53,6 +61,11 @@ def _server_action(client, *args, **kwargs):
     return client.do_action_with_server(*args, **kwargs)
 
 
+@handle_request
+def _server_clone(client, *args):
+    return client.clone_server(*args)
+
+
 @handle_request
 def _server_remove(client, *args):
     return client.delete_server(*args)
@@ -241,7 +254,13 @@ def print_servers(response: object, filters: str, ids: bool):
 @options(OUTPUT_FORMAT_OPTION)
 @click.option("--filter", "-f", "filters", default="", help="Filter output.")
 @click.option("--region", help="Use region (location).")
-@click.option("--limit", default=100, help="Limit [default: 100].")
+@click.option(
+    "--limit",
+    type=int,
+    default=500,
+    show_default=True,
+    help="Items to display.",
+)
 @click.option("--ids", is_flag=True, help="Print only server IDs.")
 def server_list(
     config, profile, verbose, region, output_format, filters, limit, ids
@@ -440,6 +459,17 @@ def get_os_id_by_name(os_images: list, os_name: str) -> int:
     return os_id
 
 
+def get_project_id_by_server_id(client, projects: list, server_id: int) -> int:
+    """Return project_id by server_id."""
+    for project in projects:
+        for server in _project_resource_list_servers(
+            client, project["id"]
+        ).json()["servers"]:
+            if server["id"] == server_id:
+                return project["id"]
+    return None
+
+
 def size_to_mb(size: str) -> int:
     """Transform string like '5G' into integer in megabytes.
     Case insensitive. For example::
@@ -495,7 +525,7 @@ def validate_bandwidth(ctx, param, value):
 
 def validate_image(client, image: str) -> int:
     """Return valid os_id or exit."""
-    log("Get list of OS images...")
+    debug("Get list of OS images...")
     os_images = _server_os_images(client).json()["servers_os"]
 
     if re.match(r"^[a-z]+-[0-9.]+$", image, re.I):
@@ -582,13 +612,13 @@ def add_ssh_key_from_file(
     # I don't want to add the same key over and over
     for exist_key in existing_ssh_keys:
         if ssh_key_body == exist_key["body"]:
-            log(
+            debug(
                 f"SSH-Key '{ssh_key_name}' already exists,"
                 f" ID {exist_key['id']} is used."
             )
             return exist_key["id"]
 
-    log(f"Add new SSH-key '{ssh_key_name}'...")
+    debug(f"Add new SSH-key '{ssh_key_name}'...")
     added_key = _ssh_key_new(
         client,
         name=ssh_key_name,
@@ -596,7 +626,7 @@ def add_ssh_key_from_file(
         is_default=False,
     )
     ssh_key_id = added_key.json()["ssh_key"]["id"]
-    log(f"New SSH-key '{ssh_key_name}' ID is '{ssh_key_id}'")
+    debug(f"New SSH-key '{ssh_key_name}' ID is '{ssh_key_id}'")
     return ssh_key_id
 
 
@@ -604,20 +634,20 @@ def add_ssh_key(client, existing_ssh_keys: list, pubkey: str) -> int:
     """Retrun SSH-key ID. from file, by SSH-key ID or by SSH-key name."""
     # From filesystem
     if os.path.exists(pubkey):
-        log(f"SSH-key to add: file: {pubkey}")
+        debug(f"SSH-key to add: file: {pubkey}")
         return add_ssh_key_from_file(client, pubkey, existing_ssh_keys)
 
     # Add by ID
     if pubkey.isdigit():
         if int(pubkey) in [key["id"] for key in existing_ssh_keys]:
-            log(f"SSH-key to add: ID: {pubkey}")
+            debug(f"SSH-key to add: ID: {pubkey}")
             return int(pubkey)
         sys.exit(f"Error: SSH-key with ID {pubkey} not found.")
 
     # Add by name
     for ssh_key in existing_ssh_keys:
         if pubkey == ssh_key["name"]:
-            log(f"SSH-key to add: name: {pubkey} ID: {ssh_key['id']}")
+            debug(f"SSH-key to add: name: {pubkey} ID: {ssh_key['id']}")
             return ssh_key["id"]
 
     sys.exit(f"Error: SSH-key '{pubkey}' not found.")
@@ -629,7 +659,9 @@ def add_ssh_key(client, existing_ssh_keys: list, pubkey: str) -> int:
 @click.option("--name", required=True, help="Cloud Server display name.")
 @click.option("--comment", help="Comment.")
 @click.option("--avatar-id", default=None, help="Avatar ID.")
-@click.option("--image", required=True, help="OS image to install.")
+@click.option(
+    "--image", required=True, help="OS ID, name or image UUID to install."
+)
 @click.option(
     "--preset-id",
     type=int,
@@ -670,6 +702,14 @@ def add_ssh_key(client, existing_ssh_keys: list, pubkey: str) -> int:
     show_default=True,
     help="Enable local network.",
 )
+@click.option(
+    "--project-id",
+    type=int,
+    default=None,
+    envvar="TWC_PROJECT",
+    callback=set_value_from_config,
+    help="Add server to specific project.",
+)
 def server_create(
     config,
     profile,
@@ -688,88 +728,108 @@ def server_create(
     ssh_key,
     ddos_protection,
     local_network,
+    project_id,
 ):
     """Create Cloud Server."""
     # pylint: disable=too-many-locals
+    # pylint: disable=too-many-branches
 
     client = create_client(config, profile)
 
+    # Set server parameters
+    payload = {
+        "name": name,
+        "comment": comment,
+        "software_id": software_id,
+        "avatar_id": avatar_id,
+        "is_ddos_guard": ddos_protection,
+        "is_local_network": local_network,
+    }
+
     # Get os_id or exit
-    log("Looking for os_id...")
+    debug("Looking for os_id...")
     os_id = validate_image(client, image)
-    log(f"os_id is {os_id}")
-    if not os_id:
-        raise click.BadParameter("Wrong image name or ID.")
+    if os_id:
+        debug(f"os_id is '{os_id}'")
+        payload["os_id"] = os_id
+    else:
+        debug("os_id not found, looking for image_id...")
+        if image in [
+            img["id"] for img in _image_list(client).json()["images"]
+        ]:
+            debug(f"image_id is '{image}'")
+            payload["image_id"] = image
+        else:
+            raise click.BadParameter("Wrong image name or ID.")
 
     # Fallback bandwidth to minimum
     if not bandwidth and not preset_id:
-        bandwidth = 100
+        payload["bandwidth"] = 100
 
     # SSH-keys
     ssh_keys_ids = []
-    log("Get SSH-keys...")
-    existing_ssh_keys = _ssh_key_list(client).json()["ssh_keys"]
+    if ssh_key:
+        debug("Get SSH-keys...")
+        existing_ssh_keys = _ssh_key_list(client).json()["ssh_keys"]
 
-    for pubkey in ssh_key:
-        ssh_keys_ids.append(add_ssh_key(client, existing_ssh_keys, pubkey))
+        for pubkey in ssh_key:
+            ssh_keys_ids.append(add_ssh_key(client, existing_ssh_keys, pubkey))
+
+    payload["ssh_keys_ids"] = ssh_keys_ids
 
     # Create Cloud Server from configurator or preset
     if cpu or ram or disk:
-        log("Get configurator...")
-        configuration = get_configuration(
+        debug("Get configurator...")
+        payload["configuration"] = get_configuration(
             client,
             DEFAULT_CONFIGURATOR_ID,
             cpu,
             ram,
             disk,
         )
-
-        # Do request
-        log("Create Cloud Server with configurator...")
-        response = _server_create(
-            client,
-            configuration=configuration,
-            os_id=os_id,
-            bandwidth=bandwidth,
-            name=name,
-            is_ddos_guard=ddos_protection,
-            is_local_network=local_network,
-            comment=comment,
-            software_id=software_id,
-            avatar_id=avatar_id,
-            ssh_keys_ids=ssh_keys_ids,
-        )
     elif preset_id:
         # Set bandwidth value from preset if option is not set
         if not bandwidth:
-            log("Check preset_id...")
+            debug("Check preset_id...")
             presets = _server_presets(client).json()["server_presets"]
             for preset in presets:
                 if preset["id"] == preset_id:
-                    log(f"Set bandwidth from preset: {preset['bandwidth']}")
-                    bandwidth = preset["bandwidth"]
+                    debug(f"Set bandwidth from preset: {preset['bandwidth']}")
+                    payload["bandwidth"] = preset["bandwidth"]
 
-        # Do request
-        log(f"Create Cloud Server with preset_id {preset_id}...")
-        response = _server_create(
-            client,
-            preset_id=preset_id,
-            os_id=os_id,
-            bandwidth=bandwidth,
-            name=name,
-            is_ddos_guard=ddos_protection,
-            is_local_network=local_network,
-            comment=comment,
-            software_id=software_id,
-            avatar_id=avatar_id,
-            ssh_keys_ids=ssh_keys_ids,
-        )
+        payload["preset_id"] = preset_id
     else:
         raise click.UsageError(
             "Configuration or preset is required. "
             "Set '--cpu', '--ram' and '--disk' or '--preset-id'"
         )
 
+    if project_id:
+        debug("Check project_id")
+        projects = _project_list(client).json()["projects"]
+        if not project_id in [prj["id"] for prj in projects]:
+            raise click.BadParameter("Wrong project ID.")
+
+    # Do request
+    debug("Create Cloud Server...")
+    response = _server_create(client, **payload)
+
+    # Add created server to project if set
+    if project_id:
+        new_server_id = response.json()["server"]["id"]
+        debug(f"Add server '{new_server_id}' to project '{project_id}'")
+        old_project_id = get_project_id_by_server_id(
+            client, projects, new_server_id
+        )
+        project_resp = _project_resource_move(
+            client,
+            from_project=old_project_id,
+            to_project=project_id,
+            resource_id=new_server_id,
+            resource_type="server",
+        )
+        debug(project_resp.text)
+
     fmt.printer(
         response,
         output_format=output_format,
@@ -803,11 +863,11 @@ def server_set_property(
     payload = {}
 
     if name:
-        payload.update({"name": name})
+        payload["name"] = name
     if comment:
-        payload.update({"comment": comment})
+        payload["comment"] = comment
     if avatar_id:
-        payload.update({"avatar_id": avatar_id})
+        payload["avatar_id"] = avatar_id
 
     if not payload:
         raise click.UsageError(
@@ -876,18 +936,18 @@ def server_resize(
     payload = {}
 
     # Save original server state
-    log("Get server original state...")
+    debug("Get server original state...")
     old_state = _server_get(client, server_id).json()["server"]
 
     # Get original server preset tags
     old_preset_tags = []
     if old_state["preset_id"]:
-        log(f"Get preset tags by preset_id {old_state['preset_id']}...")
+        debug(f"Get preset tags by preset_id {old_state['preset_id']}...")
         presets = _server_presets(client).json()["server_presets"]
         for preset in presets:
             if preset["id"] == old_state["preset_id"]:
                 old_preset_tags = preset["tags"]
-        log(f"Preset tags is {old_preset_tags}")
+        debug(f"Preset tags is {old_preset_tags}")
 
     # Return error if user tries to change dedicated server
     if "vds_dedic" in old_preset_tags:
@@ -914,7 +974,7 @@ def server_resize(
 
         # Get configurator_id if user tries to switch from preset to
         # configurator. Don't ask what is this.
-        log("Get configurator_id...")
+        debug("Get configurator_id...")
         if not configurator_id:
             if (
                 "ssd_2022" in old_preset_tags
@@ -925,14 +985,14 @@ def server_resize(
                 configurator_id = 9  # old full price configurator
 
         # Get configurator by configurator_id
-        log(f"configurator_id is {configurator_id}, get confugurator...")
+        debug(f"configurator_id is {configurator_id}, get confugurator...")
         configurators = _get_server_configurators(client).json()
         for item in configurators["server_configurators"]:
             if item["id"] == configurator_id:
                 configurator = item  # <-- this!
 
         # Check configurator and return error if configurator is unavailable
-        log(f"Configurator: '{configurator}'")
+        debug(f"Configurator: '{configurator}'")
         if configurator_id and not configurator:
             sys.exit(
                 "Error: Configurator is not available for your server. "
@@ -940,7 +1000,7 @@ def server_resize(
             )
 
         # Add configurator key to payload
-        payload.update({"configurator": {}})
+        payload["configurator"] = {}
 
         # Get original size of primary disk
         for old_disk in old_state["disks"]:
@@ -974,7 +1034,7 @@ def server_resize(
             )
 
     if bandwidth:
-        payload.update({"bandwidth": bandwidth})
+        payload["bandwidth"] = bandwidth
 
     # Handle case: user tries to change preset to another preset.
     # Check passed preset_id and exit on fail
@@ -989,7 +1049,7 @@ def server_resize(
         presets = _server_presets(client).json()["server_presets"]
         for preset in presets:
             if preset["id"] == preset_id:
-                payload.update({"preset_id": preset_id})
+                payload["preset_id"] = preset_id
         try:
             payload["preset_id"]
         except KeyError:
@@ -1014,13 +1074,33 @@ def server_resize(
 # ------------------------------------------------------------- #
 
 
+def get_ssh_keys_by_server_id(client, server_id) -> list:
+    """Return list of SSH-keys added to the server server_id."""
+    ssh_keys = _ssh_key_list(client).json()["ssh_keys"]
+    ssh_keys_list = []
+    for pubkey in ssh_keys:
+        for server in pubkey["used_by"]:
+            if server["id"] == server_id:
+                ssh_keys_list.append(pubkey["id"])
+    return ssh_keys_list
+
+
 @server.command("reinstall", help="Reinstall OS or software.")
 @options(GLOBAL_OPTIONS)
 @options(OUTPUT_FORMAT_OPTION)
-@click.option("--image", default=None, help="OS image to install.")
+@click.option(
+    "--image", default=None, help="OS ID, name or image UUID to install."
+)
 @click.option(
     "--software-id", type=int, default=None, help="Software ID to install."
 )
+@click.option(
+    "--add-ssh-keys",
+    type=bool,
+    default=True,
+    show_default=True,
+    help="Readd SSH-keys to reinstalled server.",
+)
 @click.confirmation_option(
     prompt="All data on Cloud Server will be lost.\n"
     "This action cannot be undone. Are you sure?"
@@ -1033,6 +1113,7 @@ def server_reinstall(
     output_format,
     image,
     software_id,
+    add_ssh_keys,
     server_id,
 ):
     client = create_client(config, profile)
@@ -1040,25 +1121,39 @@ def server_reinstall(
 
     if image:
         os_id = validate_image(client, image)
-        if not os_id:
-            raise click.BadParameter("Wrong image name or ID.")
-        payload.update({"os_id": os_id})
+        if os_id:
+            payload["os_id"] = os_id
+        else:
+            if image in [
+                img["id"] for img in _image_list(client).json()["images"]
+            ]:
+                payload["image_id"] = image
+            else:
+                raise click.BadParameter("Wrong image name or ID.")
+    else:
+        old_state = _server_get(client, server_id).json()["server"]
+        if old_state["image"]:
+            payload["image_id"] = old_state["image"]["id"]
+        else:
+            payload["os_id"] = old_state["os"]["id"]
 
     if software_id:
-        old_state = _server_get(client, server_id).json()["server"]
-        payload.update(
-            {
-                "os_id": old_state["os"]["id"],
-                "software_id": software_id,
-            }
-        )
+        payload["software_id"] = software_id
 
-    if not payload:
-        raise click.UsageError(
-            "Nothing to do. Set one of ['--image', '--software-id']"
-        )
+    if add_ssh_keys:
+        debug(f"Get SSH-keys by server_id '{server_id}'")
+        ssh_keys = get_ssh_keys_by_server_id(client, server_id)
+        debug(f"SSH-keys to add: {ssh_keys}")
 
+    debug(f"Reinstalling with params: {payload}")
     response = _server_update(client, server_id, payload)
+
+    if add_ssh_keys and ssh_keys:
+        debug(f"Readding SSH-keys {ssh_keys} to server '{server_id}'")
+        ssh_add_resp = _ssh_key_add(client, server_id, ssh_key_ids=ssh_keys)
+        if not ssh_add_resp.status_code == 204:
+            fmt.printer(ssh_add_resp)
+
     fmt.printer(
         response,
         output_format=output_format,
@@ -1067,10 +1162,24 @@ def server_reinstall(
 
 
 # ------------------------------------------------------------- #
-# $ twc server <action>                                         #
-# remove, boot, reboot, shutdown, clone, reset-root-password    #
+# $ twc server clone                                            #
 # ------------------------------------------------------------- #
 
+
+@server.command("clone", help="Clone Cloud Server.")
+@options(GLOBAL_OPTIONS)
+@options(OUTPUT_FORMAT_OPTION)
+@click.argument("server_id", type=int, required=True)
+def server_clone(config, profile, output_format, verbose, server_id):
+    client = create_client(config, profile)
+    response = _server_clone(client, server_id)
+    fmt.printer(
+        response,
+        output_format=output_format,
+        func=lambda response: click.echo(response.json()["server"]["id"]),
+    )
+
+
 # ------------------------------------------------------------- #
 # $ twc server remove                                           #
 # ------------------------------------------------------------- #
@@ -1093,6 +1202,11 @@ def server_remove(config, profile, verbose, server_ids):
             fmt.printer(response)
 
 
+# ------------------------------------------------------------- #
+# $ twc server <action>                                         #
+# boot, reboot, shutdown, reset-root-password                   #
+# ------------------------------------------------------------- #
+
 # ------------------------------------------------------------- #
 # $ twc server boot                                             #
 # ------------------------------------------------------------- #
@@ -1162,23 +1276,6 @@ def server_shutdown(config, profile, verbose, server_ids, hard_shutdown):
             fmt.printer(response)
 
 
-# ------------------------------------------------------------- #
-# $ twc server clone                                            #
-# ------------------------------------------------------------- #
-
-
-@server.command("clone", help="Clone Cloud Server.")
-@options(GLOBAL_OPTIONS)
-@click.argument("server_id", type=int, required=True)
-def server_clone(config, profile, verbose, server_id):
-    client = create_client(config, profile)
-    response = _server_action(client, server_id, action="clone")
-    if response.status_code == 204:
-        click.echo(server_id)
-    else:
-        fmt.printer(response)
-
-
 # ------------------------------------------------------------- #
 # $ twc server reset-root-password                              #
 # ------------------------------------------------------------- #
@@ -1216,7 +1313,7 @@ def print_presets(response: object, filters: str):
             "REGION",
             "PRICE",
             "CPU",
-            "CPU FREQ",
+            "FREQ",
             "RAM",
             "DISK",
             "DISK TYPE",
@@ -1233,8 +1330,8 @@ def print_presets(response: object, filters: str):
                 preset["price"],
                 preset["cpu"],
                 preset["cpu_frequency"],
-                preset["ram"],
-                preset["disk"],
+                str(round(preset["ram"] / 1024)) + "G",
+                str(round(preset["disk"] / 1024)) + "G",
                 preset["disk_type"],
                 preset["bandwidth"],
                 preset["description_short"],
@@ -1284,8 +1381,8 @@ def print_os_images(response: object, filters: str):
     )
     for os in os_list:
         try:
-            value = os["requirements"]["disk_min"]
-            requirements = f"disk_min: {value}G"
+            value = round(os["requirements"]["disk_min"] / 1024)
+            requirements = f"Disk >= {value}G"
         except KeyError:
             requirements = ""
         table.row(
@@ -1362,7 +1459,13 @@ def print_logs(response: object):
 @server.command("logs", help="View Cloud Server events log.")
 @options(GLOBAL_OPTIONS)
 @options(OUTPUT_FORMAT_OPTION)
-@click.option("--limit", default=100, show_default=True, help="Limit.")
+@click.option(
+    "--limit",
+    type=int,
+    default=500,
+    show_default=True,
+    help="Items to display.",
+)
 @click.option(
     "--order",
     default="asc",
@@ -1514,7 +1617,7 @@ def server_ip_add(
 ):
     client = create_client(config, profile)
     if not ipv4:
-        log("Get Cloud Server location...")
+        debug("Get Cloud Server location...")
         location = _server_get(client, server_id).json()["server"]["location"]
         if location not in REGIONS_WITH_IPV6:
             sys.exit(f"Error: IPv6 is not available in location '{location}'.")
@@ -1555,12 +1658,12 @@ def get_server_id_by_ip(client, ip_address):
 def server_ip_remove(config, profile, verbose, ip_address):
     client = create_client(config, profile)
 
-    log("Looking for IP address...")
+    debug("Looking for IP address...")
     server_id = get_server_id_by_ip(client, ip_address)
     if not server_id:
         sys.exit(f"IP address '{ip_address}' not found.")
 
-    log("Check IP...")
+    debug("Check IP...")
     ips = _server_ip_list(client, server_id).json()["server_ips"]
     for ip_addr in ips:
         if ip_addr["ip"] == ip_address and ip_addr["is_main"]:
@@ -1589,7 +1692,7 @@ def server_ip_set_ptr(
 ):
     client = create_client(config, profile)
 
-    log("Looking for IP address...")
+    debug("Looking for IP address...")
     server_id = get_server_id_by_ip(client, ip_address)
     if not server_id:
         sys.exit(f"IP address '{ip_address}' not found.")
@@ -1661,6 +1764,62 @@ def server_disk_list(config, profile, verbose, output_format, server_id):
     fmt.printer(response, output_format=output_format, func=print_disks)
 
 
+# ------------------------------------------------------------- #
+# $ twc server disk list-all                                    #
+# ------------------------------------------------------------- #
+
+
+def print_all_disks(response: object):
+    servers = response.json()["servers"]
+    table = fmt.Table()
+    table.header(
+        [
+            "SERVER",
+            "DISK",
+            "NAME",
+            "MOUNTED",
+            "SYSTEM",
+            "TYPE",
+            "STATUS",
+            "SIZE",
+            "USED",
+        ]
+    )
+    for server in servers:
+        for disk in server["disks"]:
+            table.row(
+                [
+                    server["id"],
+                    disk["id"],
+                    disk["system_name"],
+                    disk["is_mounted"],
+                    disk["is_system"],
+                    disk["type"],
+                    disk["status"],
+                    str(round(disk["size"] / 1024)) + "G",
+                    str(round(disk["used"] / 1024, 1)) + "G",
+                ]
+            )
+    table.print()
+
+
+@disk.command("list-all", aliases=["la"], help="List disks from all servers.")
+@options(GLOBAL_OPTIONS)
+@options(OUTPUT_FORMAT_OPTION)
+@click.option(
+    "--limit",
+    type=int,
+    default=500,
+    show_default=True,
+    help="Items to display."
+    " NOTE: This option affects on servers count, not disks.",
+)
+def server_disk_list_all(config, profile, verbose, output_format, limit):
+    client = create_client(config, profile)
+    response = _server_list(client, limit=limit)
+    fmt.printer(response, output_format=output_format, func=print_all_disks)
+
+
 # ------------------------------------------------------------- #
 # $ twc server disk get                                         #
 # ------------------------------------------------------------- #
@@ -1697,7 +1856,7 @@ def print_disk(response: object):
 
 
 def get_server_id_by_disk_id(client, disk_id: int) -> int:
-    log("Looking for server_id by disk_id...")
+    debug("Looking for server_id by disk_id...")
     server_id = None
     servers = _server_list(client, limit=10000).json()["servers"]
     for server in servers:
@@ -1838,7 +1997,7 @@ def print_autobackup_settings(response: object):
     type=int,
     default=1,
     show_default=True,
-    help="Number of copies to keep.",
+    help="Number of backups to keep.",
 )
 @click.option(
     "--start-date",
@@ -1857,7 +2016,7 @@ def print_autobackup_settings(response: object):
     "--day-of-week",
     type=click.IntRange(min=1, max=7),
     help="The day of the week on which backups will be created."
-    " Works only with '--interval week'.",
+    " NOTE: This option works only with interval 'week'.",
 )
 @click.argument("disk_id", type=int, required=True)
 def server_disk_autobackup(
